from dataclasses import dataclass, field
from typing import List

from pizzabot.coordinates import PizzabotCoordinate as DeltaCord


"""
Add new instructions here
"""
shorthand_description_movement = [
    ("D", "Drop pizza", DeltaCord(0, 0)),
    ("N", "Move north", DeltaCord(0, 1)),
    ("S", "Move south", DeltaCord(0, -1)),
    ("E", "Move east", DeltaCord(1, 0)),
    ("W", "Move west", DeltaCord(-1, 0)),
]


@dataclass
class Instruction:
    """
    Defines an instruction

    ...
    Attributes
    shorthand : str
        Notation printed to the CLI referencing an instruction

    description: str
        The instruction in plain text

    movement: PizzabotCoordinate
        The change in coordinates following the pizzabot running
        instruction
    """

    shorthand: str
    description: str
    movement: DeltaCord


def make_instructions():
    """
    Build instructions from list
    """
    return [Instruction(s, d, m) for s, d, m in shorthand_description_movement]


@dataclass
class PizzabotInstructions:
    """
    Defines an instructions

    ...
    Attributes
    Instructions : List
        All instructions pizzabot can run.
    """

    Instructions: List[Instruction] = field(default_factory=make_instructions)



import re

from pizzabot.grid import PizzabotGrid
from pizzabot.coordinates import PizzabotCoordinate


GRID_REGEX = r"^-?\d+x-?\d+ \("
INTEGERS_REGEX = r"-?\d+"
SPLIT_PAIRS_BRACKET = r"\("


class PizzabotInput(object):
    """
    A class to parse a string with a grid and coordinates into
    object for the pizzabot to act on

    ...

    Attributes
    ----------
    order : int
        number of expected points in a coordinate

    coordinates : list
        List of PizzabotCoordinates for the pizzabot to deliver to

    grid : PizzabotGrid
        Grid pizzabot can travel on

    Methods
    -------
    _extract_grid():
        Extracts grid from the input string

    _extract_coordinates():
        Extracts coordinates from the input string

    _parse_coordinate():
        Return pizzabot coordinate from a inter bracket string

    _validate_coordinates():
        Returns true if the number of opening and closing braces match
        the number of found coordinates, as the regexes are based on braces
    """

    def __init__(self, input_string):
        self.order = 2
        self.grid = self._extract_grid(input_string)
        self.coordinates = self._extract_coordinates(input_string)

    def _extract_grid(self, input_string):
        if re.match(GRID_REGEX, input_string) == None:
            return None
        g_strs = re.findall(INTEGERS_REGEX, input_string)[: self.order]
        g_ints = [int(x) for x in g_strs]
        return PizzabotGrid(*g_ints)

    def _extract_coordinates(self, input_string):
        coordinates = []
        c_strs = re.split(SPLIT_PAIRS_BRACKET, input_string)
        if len(c_strs) < 1:
            return []
        for c_str in c_strs[1:]:
            coordinates.append(self._parse_coordinate(c_str))
        if self._validate_coordinates(input_string, coordinates):
            return coordinates
        return []

    def _parse_coordinate(self, c_str):
        strings = re.findall(INTEGERS_REGEX, c_str)[-self.order :]
        c_int_list = [int(x) for x in strings]
        return PizzabotCoordinate(*c_int_list)

    def _validate_coordinates(self, input_string, coordinates):
        count_open_brace = 0
        count_close_brace = 0
        for letter in input_string:
            if letter == "(":
                count_open_brace += 1
            if letter == ")":
                count_close_brace += 1
        if count_open_brace == count_close_brace == len(coordinates):
            return True
        return False



from math import sqrt


class PizzabotCoordinate:
    """
    A class to represent a coordinate.

    ...

    Attributes
    ----------
    points : list
        List of integers that encode a coordinate

    Methods
    -------
    __len__():
        Returns the number of dimensions of coordinate

    _check_len():
        When preforming math on coordinate, make sure they share dimensions.

    __abs__():
        Returns the size of the coordinate

    __iadd__():
        Returns the sum of coordinates

    __sub__():
        Returns the difference of coordinates

    __sub__():
        Returns true if the coordinates are the same
    """

    def __init__(self, *argv):
        self.points = []
        for arg in argv:
            if isinstance(arg, int):
                self.points.append(arg)
            else:
                raise ValueError(
                    "ValueError exception thrown, non-int passed to PizzabotCoordinate",
                    argv,
                )

    def __len__(self):
        return len(self.points)

    def _check_len(self, other):
        if len(self) != len(other):
            raise TypeError(
                "TypeError exception thrown, PizzabotCoordinate points dimensions do not match",
                self.points,
                other.point,
            )

    def __abs__(self):
        sum_of_squares = 0
        for sp in self.points:
            sum_of_squares += sp * sp
        return sqrt(sum_of_squares)

    def __iadd__(self, other):
        added = []
        self._check_len(other)
        for sp, op in zip(self.points, other.points):
            added.append(sp + op)
        return PizzabotCoordinate(*added)

    def __sub__(self, other):
        delta = []
        self._check_len(other)
        for sp, op in zip(self.points, other.points):
            delta.append(sp - op)
        return PizzabotCoordinate(*delta)

    def __eq__(self, other):
        delta = []
        self._check_len(other)
        for sp, op in zip(self.points, other.points):
            if sp != op:
                return False
        return True



from pizzabot.input import PizzabotInput
from pizzabot.coordinates import PizzabotCoordinate
from pizzabot.instructions import PizzabotInstructions


class PizzabotEngine(object):
    """
    A class to calculate instructions from a string for the pizzabot to perform

    ...

    Attributes
    ----------
    instructions : PizzabotInstructions
        Instructions the pizzabot can perform

    position : PizzabotCoordinate
        The current position of the pizzabot

    journey : list
        Instructions performed by the pizzabot

    input : PizzabotInput
        Parses input string, to grid and coordinates objects

    Methods
    -------
    update_postion(target):
        Moves the pizzabot closer to the target by selecting the best
        instruction, records the best instruction to its journey.

    process_string(input_string):
        Takes a string from the CLI, encoding a grid and coordinates. Returns
        short hand notation for the pizzabot to deliver pizza to coordinates.
    """

    def __init__(self):
        self.instructions = PizzabotInstructions().Instructions
        self.position = PizzabotCoordinate(0, 0)
        self.journey = []
        self.input = None

    def update_postion(self, target):
        smallest_delta = abs(target - self.position)
        for instruction in self.instructions:
            delta = abs(target - instruction.movement - self.position)
            if delta < smallest_delta:
                best_instruction = instruction
                smallest_delta = delta
        self.journey.append(best_instruction)
        self.position += best_instruction.movement
        return best_instruction

    def process_string(self, input_string):
        output_string = ""
        self.input = PizzabotInput(input_string)
        if self.input.coordinates == []:
            return None
        for target in self.input.coordinates:
            while self.position != target:
                self.update_postion(target)
            self.journey.append(self.instructions[0])
        for step in self.journey:
            output_string += step.shorthand
        return output_string



#!/usr/bin/env python

from setuptools import setup

setup(name='pizzabot',
    version='1.1.3',
    author='light at zero',
    author_email='lightatzero@gmail.com',
    packages=['pizzabot'],
    install_requires=[
        "pytest",
    ],
    scripts=['bin/pizzabot'],
)
